package ibspan.ecap.merchant.shopdecisions;

import ibspan.ecap.merchant.shopdecisions.dataadministration.DataAdministrationModule;
import ibspan.ecap.merchant.shopdecisions.forecasting.ForecastDispatcher;
import ibspan.ecap.merchant.shopdecisions.forecasting.ForecastingModule;
import ibspan.ecap.merchant.shopdecisions.negotiationpreparation.NegotiationPreparingModule;
import ibspan.ecap.merchant.shopdecisions.negotiationpreparation.NegotiationTemplateDispatcher;
import ibspan.ecap.merchant.shopdecisions.negotiationpreparation.NegotiationTemplateUpdater;
import ibspan.ecap.shared.util.PropertiesProvider;
import ibspan.ecap.shared.util.TimeConverter;
import jade.content.lang.sl.SLCodec;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.CyclicBehaviour;
import jade.core.behaviours.OneShotBehaviour;
import jade.core.behaviours.ThreadedBehaviourFactory;
import jade.domain.JADEAgentManagement.JADEManagementOntology;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.lang.acl.UnreadableException;
import jade.util.Logger;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Hashtable;

//TODO: Trzyma list ostatnio wygenerowanych template'ow i na ich podstawie wypenia cen przy rejestrowaniu negocjacji
/**
 * An agent supporting decision making inside a shop. Collects information about the system 
 * events from the SA and placing them in the knowledge base. Prepares sales predictions for
 * use by the WA and creates negotiation templates and strategies for the use of the GA.
 * 
 * Uses the {@link ForecastingModule}, {@link NegotiationPreparingModule} and 
 * {@link DataAdministrationModule} to perform its goals.
 * 
 * @author Michal Drozdowicz
 * 
 */
public class ShopDecisionAgent extends Agent {
	
	protected Logger myLogger = Logger.getMyLogger(this.getClass().getName());
	
	private AID shopAgent;
	private AID warehouseAgent;
	
	private ThreadedBehaviourFactory tbf;
	private ProductProvider products = new ProductProvider();
	private PropertiesProvider propsProvider;
	private Hashtable<String, Integer> orderProblemCounters = new Hashtable<String, Integer>();
	
	private int simulationDaysLength;
	private Date simStartDate;
	private Date simEndDate;
		
	private static final String MESSAGE_KEY = "MessageKey";
	private static final String DEF_PROP_FILE_NAME = "sda.properties";
	private static final long DEF_TIME_BETWEEN_TEMPLATE_UPDATES = 5;	//in seconds
	private static final long DEF_TIME_BETWEEN_TEMPLATE_UPDATE_CYCLES = 60*60*2;	//in seconds
	private static final Date DEF_START_TIME = new Date();
	private static final double DEF_TIME_COMPRESSION = 60*12;	//a day -> 2 minutes
	private static final String DEF_WA_NAME = "wa";
	private static final String DEF_SA_NAME = "sa";
	
	private static final String START_TIME_PROP_NAME = "simulationStartDate";
	private static final String TIME_COMPRESSION_PROP_NAME = "timeCompression";
	private static final String WA_NAME_PROP_NAME = "waName";
	private static final String SA_NAME_PROP_NAME = "saName";
	private static final String TIME_BETWEEN_TEMPLATE_UPDATES_PROP_NAME = "templateUpdateSpan";
	private static final String TIME_BETWEEN_TEMPLATE_UPDATE_CYCLES_PROP_NAME = "templateUpdateCycleSpan";
	private static final String SIMULATION_DAYS_LENGTH_PROP_NAME = "simulationDaysLength";
	private static final int DEF_SIMULATION_DAYS_LENGTH = 1500;

	
	
	protected void setup() {
		myLogger.info("SDA starting");
		
		Object[] args = getArguments();
		String propFileName = "";
		for (int i = 0; i < args.length; i++) {
			propFileName += args[i]+" ";
		}
		propFileName = propFileName.trim();
		if(propFileName.length() == 0)
			propFileName = DEF_PROP_FILE_NAME;

		myLogger.fine("Property file name: "+propFileName);
				
		propsProvider = PropertiesProvider.getInstance();
		propsProvider.loadProperties(propFileName, args);
		myLogger.info("Loaded properties:\n"+propsProvider.getPropertyList());
		
		simStartDate = propsProvider.getDateProperty(START_TIME_PROP_NAME, DEF_START_TIME);
		double compressionRatio = propsProvider.getDoubleProperty(TIME_COMPRESSION_PROP_NAME, DEF_TIME_COMPRESSION);
		simulationDaysLength = propsProvider.getIntProperty(SIMULATION_DAYS_LENGTH_PROP_NAME, DEF_SIMULATION_DAYS_LENGTH);
		
		simEndDate = new Date(simStartDate.getTime()+(long)86400000*simulationDaysLength);
		myLogger.info("Simulation ends on "+new SimpleDateFormat().format(simEndDate));
		
		TimeConverter.getInstance().restart(simStartDate, compressionRatio);
		
		String waName = propsProvider.getStringProperty(WA_NAME_PROP_NAME, DEF_WA_NAME);
		String saName = propsProvider.getStringProperty(SA_NAME_PROP_NAME, DEF_SA_NAME);
		shopAgent = new AID(saName, AID.ISLOCALNAME);
		warehouseAgent = new AID(waName, AID.ISLOCALNAME);
				
		tbf = new ThreadedBehaviourFactory();
		
		//Initialising date dimension
		addBehaviour(new OneShotBehaviour(this){
			@Override
			public void action() {
				DataAdministrationModule.getInstance().initDateDimension(simStartDate, simulationDaysLength);
			}
		});
		
		
		fillBehaviors();		
	}
	
	public void endSimulation() {
//		jade.core.Runtime.instance().setCloseVM(true);
//		jade.core.Runtime.instance().shutDown();
		tbf.interrupt();
        ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
        msg.addReceiver(getAMS());
        msg.setLanguage(new SLCodec().getName());

        msg.setOntology(JADEManagementOntology.getInstance().getName()); msg.setContent("((action (agent-identifier :name "+getAMS().getName()+")(shutdown-platform)))");

        send(msg);      
        this.doDelete();
	}
	
	@Override
	protected void takeDown() {
		myLogger.info("ShopDecisionAgent dying.");
		super.takeDown();
		
	}
	
	@SuppressWarnings("serial")
	private void fillBehaviors() {
		MessageTemplate informTemplate = MessageTemplate.MatchPerformative(ACLMessage.INFORM);
		MessageTemplate saTemplate = MessageTemplate.MatchSender(shopAgent);
		MessageTemplate waTemplate = MessageTemplate.MatchSender(warehouseAgent);
		MessageTemplate contentTemplate;
		MessageTemplate template;
		
		
		//Checking for simulation end
		addBehaviour(new CyclicBehaviour(){

			@Override
			public void action() {				
				if(TimeConverter.getInstance().getDate().after(simEndDate))
				{
					myLogger.info("\n\n\n\n\n\n\nEnding simulation.\n\n\n\n\n\n");
					endSimulation();
				}
				else
				{
					
					//block for 12 hours
					block(43200000);
				}
			}
			
		});
		
		
		//TODO: Wprowadzi protokoy FIPA?
		
				
		//Wrong quantity or product run out
		contentTemplate = MessageTemplate.or(
				MessageTemplate.MatchOntology("wrongQuantity"), 
				MessageTemplate.MatchOntology("productRunOut"));
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		
		addBehaviour(new ReceivingWrapper(this, template, ProductRunOutHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new ProductRunOutHandler(myAgent);
			}
		});
				
		
		// Start forecasting
		contentTemplate = MessageTemplate.MatchOntology("startForecasting");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		addBehaviour(new ReceivingWrapper(this, template, StartForecastingHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{
				return new StartForecastingHandler(myAgent);
			}
		});
				
		
		//Stop forecasting
		contentTemplate = MessageTemplate.MatchOntology("stopForecasting");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		addBehaviour(new ReceivingWrapper(this, template, StopForecastingHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new StopForecastingHandler(myAgent);
			}
		});
		
		//Record negotiation logs
		contentTemplate = MessageTemplate.MatchOntology("negotiationLog");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		addBehaviour(new ReceivingWrapper(this, template, RecordNegotiationHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new RecordNegotiationHandler(myAgent);
			}
		});
		
		//Record transaction logs
		contentTemplate = MessageTemplate.MatchOntology("transactionLog");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		addBehaviour(new ReceivingWrapper(this, template, RecordTransactionHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new RecordTransactionHandler(myAgent);
			}
		});
		
		//Record demand data
		contentTemplate = MessageTemplate.MatchOntology("demandData");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		addBehaviour(new ReceivingWrapper(this, template, RecordDemandDataHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new RecordDemandDataHandler(myAgent);
			}
		});
		
		//Record delivery logs
		contentTemplate = MessageTemplate.MatchOntology("deliveryLog");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, waTemplate);
		addBehaviour(new ReceivingWrapper(this, template, RecordDeliveriesHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new RecordDeliveriesHandler(myAgent, orderProblemCounters);
			}
		});
		
		//Product introduced
		contentTemplate = MessageTemplate.MatchOntology("productIntroduced");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		addBehaviour(new ReceivingWrapper(this, template, MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new OneShotBehaviour(myAgent)
				{
					@Override
					public void action()
					{
						myLogger.info("Product introduced handler executing.");
						ACLMessage message = (ACLMessage) getDataStore().get(MESSAGE_KEY);
						
						if(message == null)
						{
							myLogger.warning("Product introduced handler didn't receive a message"); 
							return;
						}
						String productId;
						try
						{
							productId = (String) message.getContentObject();
						} catch (UnreadableException e)
						{
							myLogger.warning("Product introduced handler couldn't retrieve content from the message"); 
							return;
						} catch (ClassCastException e)
						{
							myLogger.warning("Product introduced handler received wrong content in the message"); 
							return;
						}				
						
						myLogger.info("SDA received info about the introduction of product: "+ productId);
						
						products.addProduct(productId);
						NegotiationPreparingModule.getInstance().productRegistered(productId);
					}
				};
			}
		});
		
		//Product withdrawn
		contentTemplate = MessageTemplate.MatchOntology("productWithdrawn");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, saTemplate);
		addBehaviour(new ReceivingWrapper(this, template, MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new OneShotBehaviour(myAgent)
				{
					@Override
					public void action()
					{
						ACLMessage message = (ACLMessage) getDataStore().get(MESSAGE_KEY);
						
						if(message == null)
						{
							myLogger.warning("Product withdrawn handler didn't receive a message"); 
							return;
						}
						String productId;
						try
						{
							productId = (String) message.getContentObject();
						} catch (UnreadableException e)
						{
							myLogger.warning("Product withdrawn handler couldn't retrieve content from the message"); 
							return;
						} catch (ClassCastException e)
						{
							myLogger.warning("Product withdrawn handler received wrong content in the message"); 
							return;
						}				
						
						products.removeProduct(productId);
						NegotiationPreparingModule.getInstance().productWithdrawn(productId);
					}
				};
			}
		});
		
		//Handle product delivery problems
		contentTemplate = MessageTemplate.MatchOntology("orderProblem");
		template = MessageTemplate.and(informTemplate, contentTemplate);
		template = MessageTemplate.and(template, waTemplate);
		addBehaviour(new ReceivingWrapper(this, template, RecordNegotiationHandler.MESSAGE_KEY)
		{
			@Override
			protected Behaviour createHandlerBehaviour()
			{				
				return new OrderProblemHandler(myAgent, orderProblemCounters);
			}
		});
		
		//Dispatch ready forecasts
		addBehaviour(tbf.wrap(new ForecastDispatcher(this, warehouseAgent)));
				
		//Periodically update product negotiation templates. Disabled for test scenarios.
		addBehaviour(new NegotiationTemplateUpdater(this, propsProvider.getLongProperty(TIME_BETWEEN_TEMPLATE_UPDATES_PROP_NAME, DEF_TIME_BETWEEN_TEMPLATE_UPDATES), propsProvider.getLongProperty(TIME_BETWEEN_TEMPLATE_UPDATE_CYCLES_PROP_NAME, DEF_TIME_BETWEEN_TEMPLATE_UPDATE_CYCLES), true));
		
		//Dispatch updated negotiation templates
		addBehaviour(tbf.wrap(new NegotiationTemplateDispatcher(this, shopAgent)));
		//Periodcally ask SA for demand information
		addBehaviour(new DemandInfoRequester(this, shopAgent, 0, 1, 0, products));
		
		//Test scenario behaviours
//		addBehaviour(new CyclicBehaviour(this)
//		{
//
//			@Override
//			public void action()
//			{
//				MessageTemplate mt = MessageTemplate.and(MessageTemplate.MatchSender(shopAgent),
//						MessageTemplate.MatchOntology("Synchronisation"));
//				ACLMessage synchroMsg = myAgent.receive(mt);
//				if(synchroMsg!=null)
//				{					
//					Date newDate; 
//					try
//					{
//						newDate = (Date) synchroMsg.getContentObject();
//					}
//					catch(ClassCastException e)
//					{
//						e.printStackTrace();
//						throw new RuntimeException("Message contained wrong content object.", e);
//					}
//					catch (UnreadableException e)
//					{
//						e.printStackTrace();
//						throw new RuntimeException("Could not read message contents.", e);
//					}
//					
//					Calendar cal = new GregorianCalendar();
//					cal.setTime(newDate);
//					cal.add(Calendar.DATE, -1);	//we're asking about yesterday
//					
//					Date yesterday = cal.getTime();
//					String productId = "Blender ZX-2010-UDPA"; //TODO: Skd bra productid 
//					ForecastingModule.getInstance().prepareForecast(productId);
//					
//					DemandRequest req = new DemandRequest();
//					req.setDate(yesterday);
//					req.setProductId(productId); 
//					
//					ACLMessage demandReq = new ACLMessage(ACLMessage.REQUEST);
//					demandReq.addReceiver(shopAgent);
//					demandReq.setOntology("DemandRequest");				
//					try
//					{
//						demandReq.setContentObject(req);
//					} catch (IOException e)
//					{				
//						e.printStackTrace();
//						throw new RuntimeException("Error setting demand data request content.",e);
//					}            
//					myAgent.send(demandReq);						
//				}
//				else
//				{
//					block();
//				}		
//			}
//			
//		});
	}

	
}
