package ibspan.ecap.merchant.shopdecisions.dataaccess;

/**
 * A class representing an observation about the sales process of a single product.  It is roughly equivalent to an entity of the Sales fact table.
 * @author  Micha Drozdowicz
 */
public class SalesFact
{
	private int negotiationEndDateKey;
	private int negotiationEndTimeKey;
	private int productOfferedKey;
	private int strategyUsedKey;
	private int templateUsedKey;
	private int negotiationId;
	private int quantityOffered;
	private float valueOffered;
	private float minimumValueOffered;
	private int quantityReserved;
	private float valueReserved;
	private int quantityConfirmed;
	private float valueConfirmed;
	private int bidCount;
	private int winningBidCount;
	private int finishedTransactionCount;
	
	/**
	 * Instantiates a new sales fact.
	 */
	public SalesFact()
	{}
	
	/**
	 * Creates a new instance of <code>SalesFact</code>. 
	 * @param negotiationEndDateKey the negotiation end date key
	 * @param negotiationEndTimeKey the negotiation end time key
	 * @param productOfferedKey the product offered key
	 * @param strategyUsedKey the strategy used key
	 * @param templateUsedKey the template used key
	 * @param negotiationId the negotiation id
	 * @param quantityOffered the quantity offered
	 * @param valueOffered the value offered
	 * @param minimumValueOffered the minimum value offered
	 * @param quantityReserved the quantity reserved
	 * @param valueReserved the value reserved
	 * @param quantityConfirmed the quantity confirmed
	 * @param valueConfirmed the value confirmed
	 * @param bidCount the bid count
	 * @param winningBidCount the winning bid count
	 * @param finishedTransactionCount the finished transaction count
	 */
	public SalesFact(int negotiationEndDateKey, int negotiationEndTimeKey,
			int productOfferedKey, int strategyUsedKey, int templateUsedKey,
			int negotiationId, int quantityOffered, float valueOffered,
			float minimumValueOffered, int quantityReserved,
			float valueReserved, int quantityConfirmed, float valueConfirmed,
			int bidCount, int winningBidCount, int finishedTransactionCount)
	{
		this.negotiationEndDateKey = negotiationEndDateKey;
		this.negotiationEndTimeKey = negotiationEndTimeKey;
		this.productOfferedKey = productOfferedKey;
		this.strategyUsedKey = strategyUsedKey;
		this.templateUsedKey = templateUsedKey;
		this.negotiationId = negotiationId;
		this.quantityOffered = quantityOffered;
		this.valueOffered = valueOffered;
		this.minimumValueOffered = minimumValueOffered;
		this.quantityReserved = quantityReserved;
		this.valueReserved = valueReserved;
		this.quantityConfirmed = quantityConfirmed;
		this.valueConfirmed = valueConfirmed;
		this.bidCount = bidCount;
		this.winningBidCount = winningBidCount;
		this.finishedTransactionCount = finishedTransactionCount;
	}
	
	/**
	 * Gets the number of bids during the negotiation.
	 *
	 * @return the number of bids during the negotiation.
	 */
	public int getBidCount()
	{
		return bidCount;
	}
	/**
	 * Sets the number of bids during the negotiation.
	 *
	 * @param bidCount the number of bids during the negotiation to set.
	 */
	public void setBidCount(int bidCount)
	{
		this.bidCount = bidCount;
	}
	/**
	 * Gets the number of finished transactions following this negotiation, that is the reservations that ended up either confirmed, cancelled or expired.
	 *
	 * @return the number of finished transactions following this negotiation
	 */
	public int getFinishedTransactionCount()
	{
		return finishedTransactionCount;
	}
	/**
	 * Sets the number of finished transactions following this negotiation, that is the reservations that ended up either confirmed, cancelled or expired.
	 *
	 * @param finishedTransactionCount the number of finished transactions following this negotiation to set
	 */
	public void setFinishedTransactionCount(int finishedTransactionCount)
	{
		this.finishedTransactionCount = finishedTransactionCount;
	}
	/**
	 * Gets the total value of the product offered using the minimum price set in the negotiation strategy.
	 *
	 * @return the total value of the product offered using the minimum price set in the negotiation strategy.
	 */
	public float getMinimumValueOffered()
	{
		return minimumValueOffered;
	}
	/**
	 * Sets the total value of the product offered using the minimum price set in the negotiation strategy.
	 *
	 * @param minimumValueOffered the the total value of the product offered using the minimum price set in the negotiation strategy to set
	 */
	public void setMinimumValueOffered(float minimumValueOffered)
	{
		this.minimumValueOffered = minimumValueOffered;
	}
	/**
	 * Gets the key to the date dimension entry describing the negotiation end date.
	 *
	 * @return the the key to the date dimension entry describing the negotiation end date.
	 */
	public int getNegotiationEndDateKey()
	{
		return negotiationEndDateKey;
	}
	/**
	 * Sets the key to the date dimension entry describing the negotiation end date.
	 *
	 * @param negotiationEndDateKey the key to the date dimension entry describing the negotiation end date.
	 */
	public void setNegotiationEndDateKey(int negotiationEndDateKey)
	{
		this.negotiationEndDateKey = negotiationEndDateKey;
	}
	/**
	 * Gets the key to the time dimension entry describing the negotiation end time.
	 *
	 * @return the key to the time dimension entry describing the negotiation end time.
	 */
	public int getNegotiationEndTimeKey()
	{
		return negotiationEndTimeKey;
	}
	/**
	 * Sets the key to the time dimension entry describing the negotiation end time.
	 *
	 * @param negotiationEndTimeKey the key to the time dimension entry describing the negotiation end time to set
	 */
	public void setNegotiationEndTimeKey(int negotiationEndTimeKey)
	{
		this.negotiationEndTimeKey = negotiationEndTimeKey;
	}
	/**
	 * Gets the negotiation's operational id.
	 *
	 * @return the negotiation's operational id.
	 */
	public int getNegotiationId()
	{
		return negotiationId;
	}
	/**
	 * Sets the negotiation's operational id.
	 *
	 * @param negotiationId the negotiation's operational id to set.
	 */
	public void setNegotiationId(int negotiationId)
	{
		this.negotiationId = negotiationId;
	}
	/**
	 * Gets the key to the product dimension entry describing the offered product.
	 *
	 * @return the key to the product dimension entry describing the offered product.
	 */
	public int getProductOfferedKey()
	{
		return productOfferedKey;
	}
	/**
	 * Sets the key to the product dimension entry describing the offered product.
	 *
	 * @param productOfferedKey the key to the product dimension entry describing the offered product to set
	 */
	public void setProductOfferedKey(int productOfferedKey)
	{
		this.productOfferedKey = productOfferedKey;
	}
	/**
	 * Gets the amount of product confirmed by the buyers and considered sold.
	 *
	 * @return the amount of product confirmed by the buyers and considered sold.
	 */
	public int getQuantityConfirmed()
	{
		return quantityConfirmed;
	}
	/**
	 * Sets the amount of product confirmed by the buyers and considered sold.
	 *
	 * @param quantityConfirmed the amount of product confirmed by the buyers and considered sold to set
	 */
	public void setQuantityConfirmed(int quantityConfirmed)
	{
		this.quantityConfirmed = quantityConfirmed;
	}
	/**
	 * Gets the amount of product offered for negotiation.
	 *
	 * @return the amount of product offered for negotiation.
	 */
	public int getQuantityOffered()
	{
		return quantityOffered;
	}
	/**
	 * Sets the amount of product offered for negotiation.
	 *
	 * @param quantityOffered the amount of product offered for negotiation.
	 */
	public void setQuantityOffered(int quantityOffered)
	{
		this.quantityOffered = quantityOffered;
	}
	/**
	 * Gets the amount of product reserved after negotiation.
	 *
	 * @return the amount of product reserved after negotiation.
	 */
	public int getQuantityReserved()
	{
		return quantityReserved;
	}
	/**
	 * Sets the amount of product reserved after negotiation.
	 *
	 * @param quantityReserved the amount of product reserved after negotiation to set
	 */
	public void setQuantityReserved(int quantityReserved)
	{
		this.quantityReserved = quantityReserved;
	}
	/**
	 * Gets the key to the strategy dimension entry describing the strategy used during the negotiation.
	 *
	 * @return the key to the strategy dimension entry describing the strategy used during the negotiation.
	 */
	public int getStrategyUsedKey()
	{
		return strategyUsedKey;
	}
	/**
	 * Sets the key to the strategy dimension entry describing the strategy used during the negotiation.
	 *
	 * @param strategyUsedKey the key to the strategy dimension entry describing the strategy used during the negotiation to set
	 */
	public void setStrategyUsedKey(int strategyUsedKey)
	{
		this.strategyUsedKey = strategyUsedKey;
	}
	/**
	 * Gets the key to the template dimension entry describing the template used during the negotiation.
	 *
	 * @return the key to the template dimension entry describing the template used during the negotiation.
	 */
	public int getTemplateUsedKey()
	{
		return templateUsedKey;
	}
	/**
	 * Sets the key to the template dimension entry describing the template used during the negotiation.
	 *
	 * @param templateUsedKey the key to the template dimension entry describing the template used during the negotiation to set
	 */
	public void setTemplateUsedKey(int templateUsedKey)
	{
		this.templateUsedKey = templateUsedKey;
	}
	/**
	 * Gets the total value of the products confirmed by the buyers and considered sold.
	 *
	 * @return the total value of the products confirmed by the buyers and considered sold.
	 */
	public float getValueConfirmed()
	{
		return valueConfirmed;
	}
	/**
	 * Sets the total value of the products confirmed by the buyers and considered sold.
	 *
	 * @param valueConfirmed the total value of the products confirmed by the buyers and considered sold to set
	 */
	public void setValueConfirmed(float valueConfirmed)
	{
		this.valueConfirmed = valueConfirmed;
	}
	/**
	 * Gets the total value of the products offered for negotiation counted using the starting price.
	 *
	 * @return the total value of the products offered for negotiation counted using the starting price.
	 */
	public float getValueOffered()
	{
		return valueOffered;
	}
	/**
	 * Sets the total value of the products offered for negotiation counted using the starting price.
	 *
	 * @param valueOffered the total value of the products offered for negotiation counted using the starting price to set
	 */
	public void setValueOffered(float valueOffered)
	{
		this.valueOffered = valueOffered;
	}
	/**
	 * Gets the total value of the products reserved after the negotiation.
	 *
	 * @return the total value of the products reserved after the negotiation.
	 */
	public float getValueReserved()
	{
		return valueReserved;
	}
	/**
	 * Sets the total value of the products reserved after the negotiation.
	 *
	 * @param valueReserved the total value of the products reserved after the negotiation to set
	 */
	public void setValueReserved(float valueReserved)
	{
		this.valueReserved = valueReserved;
	}
	/**
	 * Gets the number of winning bids.
	 *
	 * @return the number of winning bids.
	 */
	public int getWinningBidCount()
	{
		return winningBidCount;
	}
	/**
	 * Sets the number of winning bids.
	 *
	 * @param winningBidCount the number of winning bids to set
	 */
	public void setWinningBidCount(int winningBidCount)
	{
		this.winningBidCount = winningBidCount;
	}
	
	
}
