package ibspan.ecap.merchant.shopdecisions.dataadministration;

import ibspan.ecap.merchant.shopdecisions.dataaccess.SalesFact;
import ibspan.ecap.shared.ontology.negotiation.BidInfo;
import ibspan.ecap.shared.ontology.negotiation.NegotiationLog;

/**
 * A class responsible for extracting the appropriate information from the negotiation and transaction finalisation logs and creating a supplies fact object.
 * 
 * @author Micha Drozdowicz
 *
 */
public class SalesFactProcessor
{
	
	/**
	 * Creates the sales fact.
	 * 
	 * @param endDateKey the key to the date dimension entry describing the negotiation end date.
	 * @param endTimeKey the key to the time dimension entry describing the negotiation end time.
	 * @param strategyKey the key to the strategy dimension entry describing the strategy used during the negotiation.
	 * @param templateKey the key to the template dimension entry describing the template used during the negotiation.
	 * @param productKey the key to the product dimension entry describing the offered product.
	 * @param log the details of the negotiation
	 * 
	 * @return the sales fact
	 */
	public SalesFact createSalesFact(int endDateKey, int endTimeKey, 
			int strategyKey, int templateKey,
			int productKey, NegotiationLog log)
	{
		//float startingValue = log.getQuantityOffered() * log.getAskingPrice();
		//float 
		SalesFact fact = new SalesFact();
		
		fact.setNegotiationEndDateKey(endDateKey);
		fact.setNegotiationEndTimeKey(endTimeKey);
		fact.setProductOfferedKey(productKey);
		fact.setStrategyUsedKey(strategyKey);
		fact.setTemplateUsedKey(templateKey);
		fact.setNegotiationId(log.getNegotiationId());
		fact.setQuantityOffered(log.getQuantityOffered());
		fact.setMinimumValueOffered(calculateMinimumValueOffered(log));
		fact.setValueOffered(calculateValueOffered(log));
		fact.setBidCount(calculateBidCount(log));
		fact.setWinningBidCount(calculateWinningBidCount(log));
		fact.setQuantityReserved(calculateQuantityReserved(log));
		fact.setValueReserved(calculateValueReserved(log));
		fact.setValueConfirmed(0);
		fact.setQuantityConfirmed(0);
		fact.setFinishedTransactionCount(0);
		
		return fact;						
	}
	
	/**
	 * Updates the sales fact with the information available after the finalisation of a transaction.
	 * 
	 * @param fact the sales fact to update
	 * @param log the log containg information about a finished transaction
	 */
	public void updateSalesFact(SalesFact fact, TransactionLog log)
	{
		fact.setFinishedTransactionCount(fact.getFinishedTransactionCount()+1);
		if(log.getFinalisationType()==FinalisationType.CONFIRMED)
		{
			fact.setQuantityConfirmed(fact.getQuantityConfirmed()
					+ log.getQuantityReserved());
			fact.setValueConfirmed(fact.getValueConfirmed()
					+ log.getValueReserved());
		}				
	}

	private int calculateWinningBidCount(NegotiationLog log)
	{
		int winningBidCount = 0;
		if (log.getBids() != null) {
			for (BidInfo bi : log.getBids()) {
				if (bi.getReservation() != null)
					winningBidCount++;
			}
		}
		return winningBidCount;
	}

	private int calculateBidCount(NegotiationLog log)
	{
		return log.getBids() != null ? log.getBids().size() : 0;
	}

	private float calculateValueReserved(NegotiationLog log)
	{
		float value = 0;
		if (log.getBids() != null) {
			for (BidInfo bi : log.getBids()) {
				if (bi.getReservation() != null) {
					value += bi.getPrice() * bi.getQuantity();
				}
			}
		}
		return value;
	}
	
	private float calculateMinimumValueOffered(NegotiationLog log)
	{
		return log.getQuantityOffered() * log.getReservePrice();
	}

	private int calculateQuantityReserved(NegotiationLog log)
	{
		int quantity = 0;
		if (log.getBids() != null) {
			for (BidInfo bi : log.getBids()) {
				if (bi.getReservation() != null)
					quantity += bi.getQuantity();
			}
		}
		return quantity;
	}

	private float calculateValueOffered(NegotiationLog log)
	{
		return log.getQuantityOffered() * log.getAskingPrice();
	}

	
	
	
}
