package ibspan.ecap.merchant.shopdecisions.forecasting;


import ibspan.ecap.merchant.shopdecisions.ShopDecisionAgent;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.CyclicBehaviour;
import jade.core.behaviours.OneShotBehaviour;
import jade.lang.acl.ACLMessage;
import jade.util.Logger;

import java.io.IOException;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;

/**
 * A behaviour aquiring completed forecast results and dispatching them to the Warehouse  Agent. Should be run in a separate thread, because while waiting for completed tasks it blocks the running thread.
 * @author  Micha Drozdowicz
 */
public class ForecastDispatcher extends CyclicBehaviour {
	
	private ForecastingModule module; 
	private AID warehouseAgent;
	
	protected Logger myLogger = Logger.getMyLogger(this.getClass().getName()); 
	
	public ForecastDispatcher(Agent myAgent, AID warehouseAgent)
	{
		super(myAgent);
		module = ForecastingModule.getInstance();
		this.warehouseAgent = warehouseAgent;	
	}
	
	@Override
	public void action() {
		myLogger.entering(this.getClass().getName(), "action()");
		try
		{
			myLogger.fine("Taking result");
			ForecastDescription result = module.takeResult();
			if(result!=null)
			{
				myLogger.fine("Dispatching forecast");
				dispatchForecast(result);
			}
		}
		catch(InterruptedException interrupted)
		{
			myLogger.warning("Interrupted while taking forecast result.");			
			//Empty the completion queue
			pollCompletedResults();
		}
		catch(ExecutionException e){
			if(myLogger.isLoggable(Level.WARNING))
				myLogger.log(Level.WARNING, "Forecasting failed. Exception: ", e);
			final ShopDecisionAgent sda = (ShopDecisionAgent)myAgent;
			sda.addBehaviour(new OneShotBehaviour(sda)
			{

				@Override
				public void action() {
					sda.endSimulation();					
				}
				
			});
			throw new RuntimeException("Forecasting failed.",e);
		}
	}
	
	private void pollCompletedResults()
	{
		ForecastDescription result = null;
		do{				
			try
			{
				result = module.pollResult(0);
				if(result!=null)
					dispatchForecast(result);
			}
			catch(InterruptedException interrupted)
			{
				myLogger.warning("Interrupted while polling forecast result.");
			} catch (ExecutionException e) {
				if(myLogger.isLoggable(Level.WARNING))
					myLogger.log(Level.WARNING, "Forecasting failed. Exception: ", e);
			}
			}while(result!=null);
	}

	private void dispatchForecast(ForecastDescription forecast) {

		if(myLogger.isLoggable(Level.INFO))
			myLogger.info("Sending forecast to the WA. Forecast:"+forecast.toString());

		ACLMessage forecastMsg = new ACLMessage(ACLMessage.INFORM);
		forecastMsg.addReceiver(warehouseAgent);
		forecastMsg.setOntology("ForecastDescription");				
		try
		{
			forecastMsg.setContentObject(forecast);
		} catch (IOException e)
		{				
			myLogger.severe("Error setting forecast description message content.");
			throw new RuntimeException("Error setting forecast description message content.",e);
		}            
		myAgent.send(forecastMsg);
	}

}
