package ibspan.ecap.merchant.shopdecisions.forecasting;

import java.util.Date;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

/**
 * A class representing a single forecasting task, containing all the information needed to  schedule, execute at the proper time, and evaluate the forecast.  
 * @author  Micha Drozdowicz
 */
public class ForecastTask implements Callable<ForecastDescription>
{

	private static long taskCounter = 0;

	private long taskId;

	private String productId;

	private Date previousForecastEnd;

	private ForecastEvaluator evaluator;

	private Date forecastStart;

	private Future<ForecastDescription> future;

	/**
	 * Constructor
	 *
	 * @param productId the id of the product to be forecasted
	 * @param previousForecastEnd the date when the previous forecast period has finished
	 * @param evaluator the {@link ForecastEvaluator} to be used when executing the task
	 */
	public ForecastTask(String productId, Date previousForecastEnd,
			ForecastEvaluator evaluator)
	{
		this.productId = productId;
		this.previousForecastEnd = previousForecastEnd;
		this.evaluator = evaluator;

		this.taskId = getNewId();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * Uses the {@link ForecastEvaluator} passed in the constructor to prepare the forecast.
	 */
	public ForecastDescription call() throws Exception
	{
		return evaluator.prepareForecast(productId, forecastStart);
	}

	/**
	 * Gets the date at which the forecast task should be executed. 
	 *
	 * @return the date at which the forecast task should be executed.
	 */
	public Date getForecastStart()
	{
		return forecastStart;
	}
	
	/**
	 * Sets the date at which the forecast task should be executed.
	 *
	 * @param forecastStart the date at which the forecast task should be executed.
	 */
	public void setForecastStart(Date forecastStart)
	{
		this.forecastStart = forecastStart;
	}



	/**
	 * Gets the id of the product to be forecasted.
	 *
	 * @return the id of the product to be forecasted.
	 */
	public String getProductId()
	{
		return productId;
	}

	/**
	 * Gets the {@link Future} object able to return the result of the completed task.
	 *
	 * @return the {@link Future} object able to return the result of the completed task.
	 */
	public Future<ForecastDescription> getFuture()
	{
		return future;
	}

	/**
	 * Sets the {@link Future} object able to return the result of the completed task.
	 *
	 * @param future the {@link Future} object able to return the result of the completed task.
	 */	
	public void setFuture(Future<ForecastDescription> future)
	{
		this.future = future;
	}

	/**
	 * Gets the date when the previous forecast period has finished
	 *
	 * @return the date when the previous forecast period has finished
	 */
	public Date getPreviousForecastEnd()
	{
		return previousForecastEnd;
	}

	/**
	 * Gets the task's unique id
	 *
	 * @return the task's unique id
	 */
	public long getTaskId()
	{
		return taskId;
	}
	
	private static synchronized long getNewId()
	{
		long newId = ++taskCounter;
		return newId;
	}

}
