package ibspan.ecap.merchant.shopdecisions.forecasting;

import ibspan.ecap.merchant.shopdecisions.dataadministration.DataAccessException;
import ibspan.ecap.merchant.shopdecisions.dataadministration.DataAdministrationAgent;
import ibspan.ecap.merchant.shopdecisions.dataadministration.DataAdministrationModule;
import ibspan.ecap.shared.util.PropertiesProvider;
import ibspan.ecap.shared.util.TimeConverter;
import jade.util.Logger;
import java.util.Date;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;

/**
 * A subsystem responsible for preparing product sales forecasts. Designed as a singleton.
 * @author  Micha Drozdowicz
 */
public class ForecastingModule
{
	//TODO: Przy pierwszym uruchomieniu moe powinien pobra zaplanowane zadania a nastpnie je uruchomi. Tak na wypadek gdyby kto pniej zrobi persistence taskw.
	private static final ForecastingModule instance = new ForecastingModule();

	private ForecastedProductInventory inventory;

	private ForecastScheduler scheduler;

	private ScheduledForecastRunner executor;

	private ForecastEvaluator evaluator;
	private boolean recordsForecasts = true;

	protected Logger myLogger = Logger.getMyLogger(this.getClass().getName());


	private ForecastingModule()
	{		
		inventory = new ForecastedProductInventory();
		scheduler = new ForecastScheduler();
		executor = new ScheduledForecastRunner();
		evaluator = new ForecastEvaluator();
	}

	/**
	 * @return the instance of the {@linkplain ForecastingModule} class.
	 */
	public static ForecastingModule getInstance()
	{
		return instance;
	}

	/**
	 * Starts forecasting of the given product. Registers the product for periodic forecast
	 * updates and schedules an immediate preparation of the forecast.
	 *
	 * @param productId the id of the product to be forecasted.
	 */
	public void startForecasting(String productId)
	{
		myLogger.entering(getClass().getName(), "startForecasting", productId);
		if (!inventory.isForecasted(productId))
		{
			myLogger.fine("Registering product: " + productId);
			inventory.registerForForecast(productId);
			prepareForecast(productId);
		}
	}

	/**
	 * Withdraws the product from periodic sales forecasting and cancels all the scheduled
	 * forecasts.
	 *
	 * @param productId the id of the product to be withdrawn from forecast.
	 */
	public void stopForecasting(String productId)
	{
		myLogger.entering(getClass().getName(), "stopForecasting", productId);
		if (inventory.isForecasted(productId))
		{
			myLogger.fine("Withdrawing product: " + productId);
			inventory.withdrawFromForecast(productId);

			myLogger.fine("Cancelling pending tasks");
			scheduler.cancelProductForecasts(productId);
			executor.cancelProductForecasts(productId);
		}
	}

	/**
	 * Schedules an immediate preparation of a sales forecast for the specified product.
	 *
	 * @param productId the id of the product to be forecasted.
	 */
	public void prepareForecast(String productId)
	{
		myLogger.entering(getClass().getName(), "prepareForecast", productId);
		scheduleForecast(productId, TimeConverter.getInstance().getDate());
	}

	/**
	 * Performs a blocking receive of a completed forecasting task. Returns only after a task
	 * has completed or when the thread has been interrupted. 
	 *
	 * @return a forecast task result
	 * @throws InterruptedException when the thread has been interrupted while it was waiting 
	 * for a result
	 * @throws ExecutionException when an exception occurred during the calculation of the forecast.
	 */
	public ForecastDescription takeResult() throws InterruptedException,
			ExecutionException
	{
		myLogger.entering(getClass().getName(), "takeResult");
		ForecastTask task;
		try
		{
			task = executor.takeCompletedTask();
		} catch (InterruptedException stopped)
		{
			myLogger.warning("Interrupted while taking forecast future.");
			throw stopped;
		}

		ForecastDescription result = retrieveResult(task);
		forecastTaskCompleted(task, result);

		return result;
	}

	/**
	 * Performs a blocking receive of a completed forecasting task with a timeout. 
	 * Returns after a task has completed, when the timeout period specified in 
	 * <code>millisecondsTimeout</code> has elapsed or when the thread has been interrupted.
	 * 
	 * @param millisecondsTimeout the timeout after which the method should return regardless
	 * of whether a result has been completed
	 * @return a forecast task result if it has completed or <code>null</code> if the timeout has
	 * elapsed
	 * @throws InterruptedException when the thread has been interrupted while it was waiting 
	 * for a result
	 * @throws ExecutionException when an exception occurred during the calculation of the forecast.
	 */
	public ForecastDescription pollResult(long millisecondsTimeout)
			throws InterruptedException, ExecutionException
	{
		myLogger.entering(getClass().getName(), "pollResult",
				millisecondsTimeout);
		ForecastTask task;
		try
		{
			task = executor.pollForecastResult(millisecondsTimeout);
		} catch (InterruptedException stopped)
		{
			myLogger.warning("Interrupted while polling for forecast future.");
			throw stopped;
		}

		if (task != null)
		{
			ForecastDescription result = retrieveResult(task,
					millisecondsTimeout);
			forecastTaskCompleted(task, result);

			return result;
		} else
			return null;
	}

	private ForecastDescription retrieveResult(ForecastTask task)
			throws InterruptedException, ExecutionException
	{
		ForecastDescription result;
		Future<ForecastDescription> future = task.getFuture();

		if (future == null)
			return null;

		try
		{
			result = future.get();
		} catch (InterruptedException stopped)
		{
			myLogger
					.warning("Interrupted while retrieving result from future. ProductId: "
							+ task.getProductId());
			throw stopped;
		} catch (ExecutionException failed)
		{
			myLogger.warning("Forecasting of product " + task.getProductId()
					+ " failed because of execution exception: "
					+ failed.toString());
			handleExecutionFailure(task);
			throw failed;
		}

		return result;
	}

	private ForecastDescription retrieveResult(ForecastTask task,
			long millisecondsTimeout) throws InterruptedException,
			ExecutionException
	{
		ForecastDescription result = null;
		Future<ForecastDescription> future = task.getFuture();

		if (future == null)
			return null;

		try
		{
			result = future.get(millisecondsTimeout, TimeUnit.MILLISECONDS);
		} catch (TimeoutException timedOut)
		{
			return null;
		} catch (InterruptedException stopped)
		{
			myLogger
					.warning("Interrupted while retrieving result from future. ProductId: "
							+ task.getProductId());
			throw stopped;
		} catch (ExecutionException failed)
		{
			myLogger.warning("Forecasting of product " + task.getProductId()
					+ " failed because of execution exception: "
					+ failed.toString());
			handleExecutionFailure(task);
			throw failed;
		}

		return result;
	}

	protected void forecastTaskCompleted(ForecastTask task,
			ForecastDescription result) {
		scheduler.confirmTaskCompletion(task);

		if (recordsForecasts) {
			try {
				DataAdministrationModule.getInstance().recordForecastData(
						result);
			} catch (DataAccessException e) {
				myLogger.log(Level.WARNING,
						"Recording forecast details failed.", e);
			}
		}
		if (inventory.isForecasted(task.getProductId())) {
			scheduleNextForecast(result);
		}
	}

	protected void handleExecutionFailure(ForecastTask task)
	{
		if (inventory.isForecasted(task.getProductId()))
		{
			myLogger.fine("Re-running a forecast of product: "
					+ task.getProductId());
			runForecast(task);
		}
	}

	protected void scheduleNextForecast(ForecastDescription currentForecast)
	{
		myLogger.entering(getClass().getName(), "scheduleNextForecast");
		TimeConverter tc = TimeConverter.getInstance();
		Date forecastEnd = new Date(tc.getDate().getTime()
				+ currentForecast.getForecastPeriod() * 1000);
		scheduleForecast(currentForecast.getProductId(), forecastEnd);
	}

	protected void scheduleForecast(String productId, Date previousForecastEnd)
	{
		ForecastTask newTask = new ForecastTask(productId, previousForecastEnd,
				evaluator);

		synchronized (this)
		{
			ForecastTask previousTask = scheduler.getScheduledTask(productId);
			Date newDate = scheduler.evaluateNextForecastDate(productId,
					previousForecastEnd);

			if (myLogger.isLoggable(Level.FINE))
				;
			{
				if (previousTask == null)
					myLogger.fine("No previous task");
				else
					myLogger.fine("Previous task execution time: "
							+ previousTask.getForecastStart());
			}

			// Schedule a new task only if it is to be run before the one
			// already scheduled
			if (previousTask == null
					|| newDate.before(previousTask.getForecastStart()))
			{
				scheduler.scheduleForecast(newTask);
				myLogger.fine("Scheduling forecast for "
						+ newTask.getForecastStart() + ". ProductId: "
						+ productId);
				runForecast(newTask);
			}
		}
	}

	protected void runForecast(ForecastTask task)
	{
		executor.cancelProductForecasts(task.getProductId());
		executor.scheduleForecast(task);
	}

	public boolean isRecordsForecasts() {
		return recordsForecasts;
	}

	public void setRecordsForecasts(boolean recordsForecasts) {
		this.recordsForecasts = recordsForecasts;
	}
}
