package ibspan.ecap.merchant.shopdecisions.negotiationpreparation;

import ibspan.ecap.merchant.shopdecisions.dataadministration.DataAccessException;
import ibspan.ecap.merchant.shopdecisions.dataadministration.DataAdministrationModule;
import ibspan.ecap.shared.ontology.negotiation.NegotiationLog;
import ibspan.ecap.shared.util.TimeConverter;
import jade.util.Logger;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.UUID;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;

/**
 * A subsystem responsible for preparing negotiation templates and strategies.  Designed as a singleton.
 * @author  Micha Drozdowicz
 */
public class NegotiationPreparingModule
{
	private static NegotiationPreparingModule instance = new NegotiationPreparingModule();
	
	protected Logger myLogger = Logger.getMyLogger(this.getClass().getName());
	
	private NegotiationDescriptionInventory inventory;
	private PreparationTaskRunner runner;
	private TemplatePreparer preparer;
	private List<String> productsToProcess;
	private List<String> productsProcessed;
	
	private NegotiationPreparingModule()
	{
		inventory = new NegotiationDescriptionInventory();
		runner = new PreparationTaskRunner();
		preparer = new TemplatePreparer();
		productsToProcess = new LinkedList<String>();
		productsProcessed = new LinkedList<String>();
	}
	
	/**
	 * @return the instance of {@link NegotiationPreparingModule}.
	 */
	public static NegotiationPreparingModule getInstance()
	{
		return instance;
	}
	
	/**
	 * Performs a blocking receive of a completed task. Returns only after a task
	 * has completed or when the thread has been interrupted. 
	 *
	 * @return a complete {@link NegotiationDescription} result
	 * @throws InterruptedException when the thread has been interrupted while it was waiting 
	 * for a result
	 * @throws ExecutionException when an exception occurred during the preparation of the negotiation template
	 */
	public NegotiationDescription takeResult() throws InterruptedException, ExecutionException
	{
		NegotiationDescription result = runner.takeResult();
		//record an empty negotiation log to record the price and amount offered
		NegotiationLog log = new NegotiationLog();
		log.setAskingPrice(result.getStartingPrice());
		log.setNegotiationStart(TimeConverter.getInstance().getDate());
		log.setNegotiationEnd(TimeConverter.getInstance().getDate());
		log.setNegotiationId(Math.abs((int) UUID.randomUUID()
				.getLeastSignificantBits()));
		log.setProductId(result.getProductId());
		log.setQuantityOffered(result.getOfferedAmount());
		log.setReservePrice(result.getReservePrice());
		log.setTemplateId(result.getTemplateId());
		log.setStrategyId(result.getStrategyId());
		try {
			DataAdministrationModule.getInstance().recordNegotiationDetails(log);
		} catch (DataAccessException e) {
			myLogger.log(Level.WARNING, "Error recording information about the negotiation description.", e);
		}
		if(productsProcessed.contains(result.getProductId()) || productsToProcess.contains(result.getProductId()))
		{
			inventory.setLastDescription(result.getProductId(), result);
		}
		return result;
	}
	
	/**
	 * Synchronously prepares a new negotiation template for the given product.
	 *
	 * @param productId the id of the product for which a negotiation template is to be prepared.
	 * @return a new negotiation template.
	 */
	public NegotiationDescription prepareNegotiationDescription(String productId)
	{
		NegotiationDescription previousDescription = inventory.getLastDescription(productId);
		
		NegotiationDescription result = preparer.prepareTemplate(productId, TimeConverter.getInstance().getDate(), previousDescription); 
				
		if(productsProcessed.contains(result.getProductId()) || productsToProcess.contains(result.getProductId()))
		{
			inventory.setLastDescription(result.getProductId(), result);
		}
		return result;
	}
	
	/**
	 * Requests the preparation of a negotiation template for the given product. This operation
	 * runs asynchronously and the result can be obtained using the {@link #takeResult} method.	 *  
	 *
	 * @param productId the product whose template should be prepared
	 */
	public void requestNegotiationDescription(String productId)
	{
		NegotiationDescription previousDescription = inventory.getLastDescription(productId);
		
		runner.runPreparationTask(new PreparationTask(productId, TimeConverter.getInstance().getDate(), previousDescription, preparer));
	}
	
	/**
	 * Requests the preparation of a negotiation template for the next registered product. 
	 * This operation runs asynchronously and the result can be obtained using 
	 * the {@link #takeResult} method.  
	 *
	 */
	public void requestNextNegotiationDescription()
	{
		String productId = getNextProductId();
		if(productId!=null)
			requestNegotiationDescription(productId);
	}
	
	/**
	 * Registers the product for periodic template updates and requests an immediate
	 * asynchronous preparation of the first template.
	 *
	 * @param productId the id of the product to be handled
	 */
	public void productRegistered(String productId)
	{
		if(!productsProcessed.contains(productId) && !productsToProcess.contains(productId))
		{
			productsToProcess.add(productId);
			inventory.registerProduct(productId);
			
			requestNegotiationDescription(productId);
		}
	}
	
	/**
	 * Withdraws the product from periodic template updates.
	 *
	 * @param productId the id of the product to be withdrawn
	 */
	public void productWithdrawn(String productId)
	{
		productsToProcess.remove(productId);
		productsProcessed.remove(productId);
		inventory.withdrawProduct(productId);
	}
	
	private String getNextProductId()
	{
		if (productsToProcess.size() == 0)
		{
			if(productsProcessed.size() == 0)
			{
				return null;	//nothing to process
			}
			else
			{
				productsToProcess = new LinkedList(productsProcessed);
				productsProcessed.clear();
			}
		}
		
		// TODO: Synchronisation needed?
		String productId = productsToProcess.get(0);
		productsToProcess.remove(0);
		productsProcessed.add(productId);
		return productId;

	}

	public boolean roundFinished()
	{
		return productsToProcess.size() == 0;
	}
}
