package ibspan.ecap.merchant.shopdecisions.negotiationpreparation;

import ibspan.ecap.merchant.shopdecisions.dataaccess.DataQueryService;
import ibspan.ecap.merchant.shopdecisions.dataaccess.ResultProcessor;
import ibspan.ecap.merchant.shopdecisions.dataadministration.DataAccessException;
import ibspan.ecap.shared.util.DateTool;
import ibspan.ecap.shared.util.PropertiesProvider;
import ibspan.ecap.shared.util.TimeConverter;

import jade.util.Logger;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Random;
import java.util.logging.Level;

import FIPA.DateTime;

/**
 * A class preparing the negotiation templates.
 *
 * @author Micha Drozdowicz
 *
 */
public class TemplatePreparer
{
	public class AmountInputData
	{
		private float movingAverageAmountOffered = -1;

		public float getMovingAverageAmountOffered()
		{
			return movingAverageAmountOffered;
		}

		public void setMovingAverageAmountOffered(float averageQuantityReserved)
		{
			this.movingAverageAmountOffered = averageQuantityReserved;
		}
	}

	public class PriceInputData
	{
		private int lastAmountSold = -1;
		private float lastAmountOffered = -1;
		private float lastValueSold = -1;
		private float lastReserveValueOffered = -1;
		private int lastClientsRegistered = -1;
		private int beforeLastAmountSold = -1;
		private float beforeLastAmountOffered = -1;
		private float beforeLastValueSold = -1;		
		private float beforeLastReserveValueOffered = -1;
		private int beforeLastClientsRegistered = -1;
		private int periodNumber = -1;		 		
		
		public int getLastAmountSold()
		{
			return lastAmountSold;
		}
		public void setLastAmountSold(int lastAmountSold)
		{
			this.lastAmountSold = lastAmountSold;
		}
		public float getLastValueSold()
		{
			return lastValueSold;
		}
		public void setLastValueSold(float lastValueSold)
		{
			this.lastValueSold = lastValueSold;
		}
		public int getBeforeLastAmountSold()
		{
			return beforeLastAmountSold;
		}
		public void setBeforeLastAmountSold(int beforeLastAmountSold)
		{
			this.beforeLastAmountSold = beforeLastAmountSold;
		}
		public float getBeforeLastValueSold()
		{
			return beforeLastValueSold;
		}
		public void setBeforeLastValueSold(float beforeLastValueSold)
		{
			this.beforeLastValueSold = beforeLastValueSold;
		}
		public int getLastClientsRegistered()
		{
			return lastClientsRegistered;
		}
		public void setLastClientsRegistered(int lastClientsRegistered)
		{
			this.lastClientsRegistered = lastClientsRegistered;
		}
		public int getBeforeLastClientsRegistered()
		{
			return beforeLastClientsRegistered;
		}
		public void setBeforeLastClientsRegistered(int beforeLastClientsRegistered)
		{
			this.beforeLastClientsRegistered = beforeLastClientsRegistered;
		}
		public int getPeriodNumber()
		{
			return periodNumber;
		}
		public void setPeriodNumber(int periodNumber)
		{
			this.periodNumber = periodNumber;
		}
		public float getLastReserveValueOffered()
		{
			return lastReserveValueOffered;
		}
		public void setLastReserveValueOffered(float lastReserveValueOffered)
		{
			this.lastReserveValueOffered = lastReserveValueOffered;
		}
		public float getBeforeLastReserveValueOffered()
		{
			return beforeLastReserveValueOffered;
		}
		public void setBeforeLastReserveValueOffered(float beforeLastReserveValueOffered)
		{
			this.beforeLastReserveValueOffered = beforeLastReserveValueOffered;
		}
		public float getLastAmountOffered()
		{
			return lastAmountOffered;
		}
		public void setLastAmountOffered(float lastAmountOffered)
		{
			this.lastAmountOffered = lastAmountOffered;
		}
		public float getBeforeLastAmountOffered()
		{
			return beforeLastAmountOffered;
		}
		public void setBeforeLastAmountOffered(float beforeLastAmountOffered)
		{
			this.beforeLastAmountOffered = beforeLastAmountOffered;
		}		
	}

	private static final double DEF_BASE_PRICE_MODIFIER = 1;
	private static final String BASE_PRICE_MODIFIER_PROP_NAME = "basePriceModifier";
	private static final double DEF_MAX_BASE_PRICE = 10;
	private static final String MAX_BASE_PRICE_PROP_NAME = "maxBasePrice";
	private static final String PRICE_MODIFIER_FRACTION_LIMIT_PROP_NAME = "priceModifierFractionLimit";
	private static final String AMOUNT_OFFERED_MARGIN_FACTOR_PROP_NAME = "amountOfferedMarginFactor";
	//The number of days based on which the amount of products offered is estimated
	private static final String AMOUNT_OFFERED_AVERAGE_DAYS_PROP_NAME = "amountOfferedMovingAverageDays";
	private static final double DEF_PRICE_MODIFIER_FRACTION_LIMIT = 30;
	private static final int DEF_AMOUNT_OFFERED_AVERAGE_DAYS = 5;
	//The fraction of the predicted sales amount to be added as a margin.
	private static final double DEF_AMOUNT_OFFERED_MARGIN_FACTOR = 0.1;
	private static final String MIN_AMOUNT_OFFERED_PROP_NAME = "minAmountOffered";
	private static final int DEF_MIN_AMOUNT_OFFERED = 5;
	private static final String DEFAULT_TEMPLATE_ID_PROP_NAME = "defaultTemplate";
	private static final String DEF_DEFAULT_TEMPLATE_ID= "English auction";
	private static final String DEFAULT_STRATEGY_ID_PROP_NAME = "minAmountOffered";
	private static final String DEF_DEFAULT_STRATEGY_ID = "Simple strategy";
	private static final String MIN_PRICE_PROP_NAME = "minPrice";
	private static final double DEF_MIN_PRICE = 3.5;
	
	
	
	private double basePriceModifier;
	private double maxBasePrice;
	private double priceModifierFractionLimit;
	private int amountOfferedMovingAverageDays;	
	private double amountOfferedMarginFactor; 
	private int minAmount;
	private double minPrice;
	private Random random = new Random(new Date().getTime());
	
	private Logger myLogger = Logger.getMyLogger(this.getClass().getName());

	/**
	 * Prepares a negotiation template for a given product.
	 *
	 * @param productId the id of the product for which a negotiation template should be prepared.
	 * @param date the date starting at which the template is to be used. 
	 * @param previousTemplate the previously used template.
	 * @return the updated negotiation template.
	 */
	public NegotiationDescription prepareTemplate(String productId, Date date, NegotiationDescription previousTemplate)
	{
		myLogger.entering("Entering prepareTemplate", new SimpleDateFormat().format(date));
		PropertiesProvider propsProvider = PropertiesProvider.getInstance();
		minPrice = propsProvider.getDoubleProperty(MIN_PRICE_PROP_NAME, DEF_MIN_PRICE);
		basePriceModifier = propsProvider.getDoubleProperty(BASE_PRICE_MODIFIER_PROP_NAME, DEF_BASE_PRICE_MODIFIER);
		maxBasePrice = propsProvider.getDoubleProperty(MAX_BASE_PRICE_PROP_NAME, DEF_MAX_BASE_PRICE);
		priceModifierFractionLimit = propsProvider.getDoubleProperty(PRICE_MODIFIER_FRACTION_LIMIT_PROP_NAME, DEF_PRICE_MODIFIER_FRACTION_LIMIT);
		amountOfferedMovingAverageDays = propsProvider.getIntProperty(AMOUNT_OFFERED_AVERAGE_DAYS_PROP_NAME, DEF_AMOUNT_OFFERED_AVERAGE_DAYS);
		amountOfferedMarginFactor = propsProvider.getDoubleProperty(AMOUNT_OFFERED_MARGIN_FACTOR_PROP_NAME, DEF_AMOUNT_OFFERED_MARGIN_FACTOR);
		minAmount = propsProvider.getIntProperty(MIN_AMOUNT_OFFERED_PROP_NAME, DEF_MIN_AMOUNT_OFFERED);
		String strategyId = propsProvider.getStringProperty(DEFAULT_STRATEGY_ID_PROP_NAME, DEF_DEFAULT_STRATEGY_ID);
		String templateId = propsProvider.getStringProperty(DEFAULT_TEMPLATE_ID_PROP_NAME, DEF_DEFAULT_TEMPLATE_ID);
		
		myLogger.info("MinAmount:"+minAmount);
		myLogger.info("MinPrice:"+minPrice);
		myLogger.info("MaxBasePrice:"+maxBasePrice);
		
		float startingPrice;
		try
		{
			startingPrice = (float) Math.max(minPrice, evaluateStartingPrice(productId, date, previousTemplate));
		} catch (DataAccessException e)
		{
			myLogger.log(Logger.SEVERE, "Failed to access data while evaluating the price of the product.", e);
			throw new RuntimeException("Failed to access data while preparing negotiation template.");
		}
		float reservePrice = startingPrice;
		int amountOffered;
		try
		{
			amountOffered = Math.max(minAmount, evaluateAmountOffered(productId, date, previousTemplate));
		} catch (DataAccessException e)
		{
			myLogger.log(Logger.SEVERE, "Failed to access data while evaluating the amount of offered products.", e);
			throw new RuntimeException("Failed to access data while preparing negotiation template.");
		}
		return new NegotiationDescription(TimeConverter.getInstance().getDate(), productId, startingPrice, reservePrice, amountOffered, strategyId, templateId);
	}

	private int evaluateAmountOffered(String productId, Date date, NegotiationDescription previousTemplate) throws DataAccessException
	{	
		AmountInputData data;
		try
		{
			data = fetchAmountInputData(productId, date);
		} catch (SQLException e)
		{
			myLogger.log(Logger.FINE, "SqlException occurred while fetching data from database.", e);
			throw new DataAccessException("Failed querying the data source.");
		}
		
		double amount = data.getMovingAverageAmountOffered();
		amount += amount * amountOfferedMarginFactor;
		return (int) Math.ceil(amount);
	}

	
	public double evaluateStartingPrice(String productId, Date date, NegotiationDescription previousTemplate) throws DataAccessException
	{
		PriceInputData data;
		try
		{
			data = fetchPriceInputData(productId, date);
		} catch (SQLException e)
		{
			e.printStackTrace();
			throw new DataAccessException("Failed querying the data source.");
		}
		
		double n0 = data.periodNumber / priceModifierFractionLimit;		
		double priceModifier = basePriceModifier * (n0 + 1)/(n0 + data.periodNumber);
		
		double price;	
		double lastPrice=0;
		double beforeLastPrice;
				
		if(data.lastAmountSold < 0 || data.lastClientsRegistered < 0 || data.lastValueSold < 0)
		{
			//no data about last period
			price = random.nextDouble()*maxBasePrice;
		}
		else if(data.beforeLastAmountSold < 0 || data.beforeLastClientsRegistered < 0 || data.beforeLastValueSold < 0)
		{
			//no data about the period before last
			lastPrice = data.lastReserveValueOffered / data.lastAmountOffered;
						
			if(random.nextBoolean())	//randomly decide to raise or lower the price
				priceModifier *= -1;	
			
			price = lastPrice * (1 + priceModifier);
		}
		else //data about 2 previous periods
		{
			lastPrice = data.lastReserveValueOffered / data.lastAmountOffered;
			beforeLastPrice = data.beforeLastReserveValueOffered / data.beforeLastAmountOffered;
			if(lastPrice < beforeLastPrice)
				priceModifier *= -1;			//previous modifier was negative
						
			float beforeLast; 
			float last;
			
			if(data.beforeLastClientsRegistered>0 && data.lastClientsRegistered>0)
			{
				beforeLast = data.beforeLastValueSold / data.beforeLastClientsRegistered;
				last = data.lastValueSold / data.lastClientsRegistered;
			}
			else
			{
				beforeLast = data.beforeLastValueSold;
				last = data.lastValueSold;
			}
			
			if(last < beforeLast)
				priceModifier *= -1;			//previous modifier didn't help - need to reverse it
						
			price = lastPrice * (1 + priceModifier);
		}
		
		myLogger.fine("Evaluating price. n0= "+n0+"; periodNumber = "+data.periodNumber+"; priceModifier = "+priceModifier+";lastPrice="+lastPrice+"; price = "+price);
		return price;
	}
	
	private AmountInputData fetchAmountInputData(String productId, Date date) throws SQLException
	{
		final Calendar cal = new GregorianCalendar();
		cal.setTime(date);
		cal.add(Calendar.DATE, -1);
		final Date averagePeriodEnd = cal.getTime();
		cal.add(Calendar.DATE, -amountOfferedMovingAverageDays);
		final Date averagePeriodStart = cal.getTime();

		DateFormat df = DateFormat.getDateInstance();

		final AmountInputData data = new AmountInputData();

		DataQueryService dqs = DataQueryService.getInstance();
		String salesQuery = "SELECT AVG(fsa_quantityreserved) as avgquantityreserved FROM ft_sales " +
				"INNER JOIN dim_date ON fsa_ddt_negoenddateid = ddt_id "
				+ "INNER JOIN dim_product ON fsa_dpd_productid = dpd_id WHERE dpd_globalproductid='"
				+ productId
				+ "' AND ddt_fulldate > '"
				+ df.format(averagePeriodStart)
				+ "' AND ddt_fulldate <= '" + df.format(averagePeriodEnd) + "' AND fsa_bidcount>0 ";
		myLogger.finer("Executing quantity reserved query. Query content: " + salesQuery);
		dqs.executeQuery(salesQuery, new ResultProcessor()
		{
			public void processResult(ResultSet rs) throws SQLException
			{
				if(rs.next())
				{
					float averageQuantityReserved = rs.getFloat("avgquantityreserved");
					
					data.setMovingAverageAmountOffered(averageQuantityReserved);
				}
			}
		});
		return data;
	}

	private PriceInputData fetchPriceInputData(String productId, Date date) throws SQLException
	{
		final Calendar cal = new GregorianCalendar();
		cal.setTime(date);
		cal.add(Calendar.DATE, -1);
		final Date last = cal.getTime();
		cal.add(Calendar.DATE, -1);
		final Date beforeLast = cal.getTime();

		DateFormat df = DateFormat.getDateInstance();

		final PriceInputData data = new PriceInputData();

		DataQueryService dqs = DataQueryService.getInstance();

		StringBuilder sb = new StringBuilder();
		sb.append("SELECT sales.ddt_dayinepoch, amountsold, amountoffered, valuesold, reservevalue, fdm_registeredclients as registeredclients FROM ");
		sb.append(" (SELECT ddt_dayinepoch, dpd_id, SUM(fsa_quantityreserved) as amountsold, ");
		sb.append("		SUM(fsa_quantityoffered) as amountoffered, SUM(fsa_valuereserved) as valuesold, ");
		sb.append("		SUM(fsa_minvalueoffered) as reservevalue FROM ft_sales ");
		sb.append("		RIGHT OUTER JOIN dim_date ON ddt_id = fsa_ddt_negoenddateid ");
		sb.append("		RIGHT OUTER JOIN dim_product ON fsa_dpd_productid = dpd_id ");
		sb.append("		WHERE dpd_globalproductid='"+productId+"' AND ddt_fulldate < '"+df.format(date)+"' AND fsa_bidcount>0 ");
		sb.append("		GROUP BY ddt_dayinepoch, dpd_id ");
		sb.append("		ORDER BY ddt_dayinepoch DESC");
		sb.append("		LIMIT 2 ");
		sb.append(") as sales ");
		sb.append("LEFT OUTER JOIN ");
		sb.append("(SELECT ddt_dayinepoch, dpd_id, fdm_registeredclients ");
		sb.append("		FROM ft_demand ");
		sb.append("		RIGHT OUTER JOIN dim_date ON ddt_id = fdm_ddt_dateid ");
		sb.append("		RIGHT OUTER JOIN dim_product ON fdm_dpd_productid = dpd_id ");
		sb.append("		WHERE dpd_globalproductid='"+productId+"' AND ddt_fulldate < '"+df.format(date)+"' ");
		sb.append(") as demand ");
		sb.append("ON sales.ddt_dayinepoch = demand.ddt_dayinepoch AND sales.dpd_id = demand.dpd_id ");

		String salesQuery = sb.toString();
			
		myLogger.finer("Executing sales query. Query content: " + salesQuery);
		dqs.executeQuery(salesQuery, new ResultProcessor()
		{
			public void processResult(ResultSet rs) throws SQLException
			{
				if(rs.next())
				{
					//Last recorded period
					data.lastAmountSold = rs.getInt("amountsold");
					data.lastAmountOffered = rs.getInt("amountoffered");
					data.lastValueSold = rs.getFloat("valuesold");
					data.lastReserveValueOffered = rs
							.getFloat("reservevalue");
					data.lastClientsRegistered = rs
							.getInt("registeredclients");
				}
				if(rs.next())
				{
					//Previous to last
					data.beforeLastAmountSold = rs.getInt("amountsold");
					data.beforeLastAmountOffered = rs
							.getInt("amountoffered");
					data.beforeLastValueSold = rs.getFloat("valuesold");
					data.beforeLastReserveValueOffered = rs
							.getFloat("reservevalue");
					data.beforeLastClientsRegistered = rs
							.getInt("registeredclients");
				}			
			}
		});
		String periodQuery = "SELECT MIN(ddt_dayinepoch) as introductionDay"
				+ "	FROM ft_sales INNER JOIN dim_product "
				+ "		ON fsa_dpd_productid = dpd_id"
				+ "	INNER JOIN dim_date" + "		ON fsa_ddt_negoenddateid = ddt_id"
				+ "	WHERE dpd_globalproductid='" + productId
				+ "'	AND ddt_fulldate <= '" + df.format(last)+"'";
		myLogger.finer("Executing period query. Query content: " + periodQuery);
		final int dayInEpoch = DateTool.getDayInEpoch(date);
		dqs.executeQuery(periodQuery, new ResultProcessor()
		{
			public void processResult(ResultSet rs) throws SQLException
			{
				while (rs.next())
				{
					int introDay = rs.getInt("introductionDay");
					
					data.setPeriodNumber(dayInEpoch - introDay);
				}
			}
		});

		return data;
	}
}
