package ibspan.ecap.merchant.stubs;

import java.io.IOException;
import java.util.Date;

import ibspan.ecap.merchant.shopdecisions.dataadministration.DeliveryLog;
import ibspan.ecap.merchant.shopdecisions.forecasting.ForecastDescription;
import ibspan.ecap.merchant.shopdecisions.negotiationpreparation.NegotiationDescription;
import ibspan.ecap.shared.util.PropertiesProvider;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.CyclicBehaviour;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.lang.acl.UnreadableException;
import jade.util.Logger;

/**
 * An agent mimicking the behaviour of the WarehouseAgent in interactions with the ShopDecisionAgent.
 * 
 * @author Michal Drozdowicz
 *
 */
public class WarehouseAgentStub extends Agent
{
	public class WAStubBehaviour extends CyclicBehaviour
	{

		private static final String ORDER_DELAY_PROP_NAME = "orderDelay";
		private static final long DEF_ORDER_DELAY = 1000*60*10;
		private static final String WHOLESALER_ID_PROP_NAME = "wholesalerId";
		private static final String DELIVERY_DELAY_PROP_NAME = "deliveryDelay";
		private static final long DEF_DELIVERY_DELAY = 1000*60*60*10;
		private static final String DEF_WHOLESALER_ID = "Wholesaler1";

		@Override
		public void action()
		{
			PropertiesProvider propsProvider = PropertiesProvider.getInstance();
			long orderDelay = propsProvider.getLongProperty(ORDER_DELAY_PROP_NAME, DEF_ORDER_DELAY);
			long deliveryDelay = propsProvider.getLongProperty(DELIVERY_DELAY_PROP_NAME, DEF_DELIVERY_DELAY);
			String wholesalerId = propsProvider.getStringProperty(WHOLESALER_ID_PROP_NAME, DEF_WHOLESALER_ID);
			
			myLogger.info("Warehouse stub waiting for prediction.");
			MessageTemplate mt = MessageTemplate.MatchSender(sda);
			ACLMessage paramsMsg = myAgent.receive(mt);
			if(paramsMsg!=null)
			{				
				ForecastDescription descr;
				try
				{
					descr = (ForecastDescription) paramsMsg.getContentObject();
				}
				catch(ClassCastException e)
				{
					e.printStackTrace();
					throw new RuntimeException("Message contained wrong content object.", e);
				}
				catch (UnreadableException e)
				{
					e.printStackTrace();
					throw new RuntimeException("Could not read message contents.", e);
				}
				
				myLogger.info("Warehouse stub received prediction. Sending delivery log.");
				
				DeliveryLog log = new DeliveryLog();
				log.setDeliveredProductId(descr.getProductId());
				log.setOrderTime(new Date(descr.getPreparationTime().getTime()+orderDelay));
				log.setDeliveryTime(new Date(log.getOrderTime().getTime()+deliveryDelay));
				log.setOrderedQuantity((int) (descr.getForecastAmount()+descr.getForecastDeviation()));
				log.setWholesalerId(wholesalerId);
				
				ACLMessage logMsg = new ACLMessage(ACLMessage.INFORM);
				logMsg.addReceiver(sda);
				try
				{
					logMsg.setContentObject(log);
				} catch (IOException e)
				{				
					e.printStackTrace();
					throw new RuntimeException("Error setting delivery log message content.",e);
				}            
				myAgent.send(logMsg);	
				myLogger.info("Warehouse stub sent the delivery log.");
			}
			else
			{
				block();
			}
		}
		
	}
	
	private static final String DEF_PROP_FILE_NAME = "wastub.properties";
	private static final String DEF_SDA_NAME = "sda";
	
	private static final String SDA_PROP_NAME = "sdaName";
	
	private AID sda;
	private Logger myLogger = Logger.getMyLogger(this.getClass().getName());
	
	@Override
	protected void setup()
	{
		super.setup();
		
		Object[] args = getArguments();
		String propFileName = "";
		for (int i = 0; i < args.length; i++) {
			propFileName += args[i]+" ";
		}
		propFileName = propFileName.trim();
		if(propFileName.length() == 0)
			propFileName = DEF_PROP_FILE_NAME;
//		String propFileName = (args!=null && args.length > 0) ? args[0].toString() : DEF_PROP_FILE_NAME;

		
		myLogger.fine("Property file name: "+propFileName);
		
		PropertiesProvider propsProvider = PropertiesProvider.getInstance();
		propsProvider.loadProperties(propFileName, args);
		myLogger.info("Loaded properties:\n"+propsProvider.getPropertyList());
		
		String sdaName = propsProvider.getStringProperty(SDA_PROP_NAME, DEF_SDA_NAME);
		sda = new AID(sdaName, AID.ISLOCALNAME);
		
		addBehaviour(new WAStubBehaviour());
	}
}
