package ibspan.ecap.shared.util;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.Delayed;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Future;
import java.util.concurrent.FutureTask;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

/**
 * A {@link CompletionService} that uses a supplied
 * {@link ScheduledExecutorService} to schedule and execute tasks. This class
 * arranges that submitted tasks are, upon completion, placed on a queue
 * accessible using <tt>take</tt>. This class is based on the
 * {@link ExecutorCompletionService}.
 * 
 * @author Micha Drozdowicz
 * 
 * @param <RES>
 *            the type of the results returned by the executed tasks
 */
public class ScheduledExecutorCompletionService<RES> implements
		CompletionService<RES> {
	private ScheduledExecutorService scheduler;
	private BlockingQueue<Future<RES>> completionQueue;

	/** Base of nanosecond timings, to avoid wrapping */
	private static final long NANO_ORIGIN = System.nanoTime();

	/**
	 * Returns nanosecond time offset by origin
	 */
	final long now() {
		return System.nanoTime() - NANO_ORIGIN;
	}

	/**
	 * FutureTask extension to enqueue upon completion
	 */
	private class ScheduledQueueingFuture extends FutureTask<RES> implements
			ScheduledFuture<RES> {
		/** The time the task is enabled to execute in nanoTime units */
		private long time = 0;

		ScheduledQueueingFuture(Callable<RES> c) {
			super(c);
		}

		ScheduledQueueingFuture(Runnable t, RES r) {
			super(t, r);
		}

		protected void done() {
			if (!isCancelled()) {
				completionQueue.add(this);
			}
		}

		void setDelay(long ns) {
			time = now() + ns;
		}

		public long getDelay(TimeUnit unit) {
			long d = unit.convert(time - now(), TimeUnit.NANOSECONDS);
			return d;
		}

		public int compareTo(Delayed other) {
			long diff = getDelay(TimeUnit.NANOSECONDS)
					- other.getDelay(TimeUnit.NANOSECONDS);
			if (diff < 0)
				return -1;
			else if (diff > 0)
				return 1;
			else
				return 0;
		}
	}

	/**
	 * Creates a ScheduledExecutorCompletionService using the supplied scheduler
	 * for base task execution and a {@link LinkedBlockingQueue} as a completion
	 * queue.
	 * 
	 * @param scheduler
	 *            the scheduler to use
	 * @throws NullPointerException
	 *             if scheduler is <tt>null</tt>
	 */
	public ScheduledExecutorCompletionService(ScheduledExecutorService scheduler) {
		if (scheduler == null)
			throw new NullPointerException();
		this.scheduler = scheduler;
		this.completionQueue = new LinkedBlockingQueue<Future<RES>>();
	}

	/**
	 * Creates a ScheduledExecutorCompletionService using the supplied scheduler
	 * for base task execution and the supplied queue as its completion queue.
	 * 
	 * @param scheduler
	 *            the scheduler to use
	 * @param completionQueue
	 *            the queue to use as the completion queue normally one
	 *            dedicated for use by this service
	 * @throws NullPointerException
	 *             if scheduler or completionQueue are <tt>null</tt>
	 */
	public ScheduledExecutorCompletionService(
			ScheduledExecutorService scheduler,
			BlockingQueue<Future<RES>> completionQueue) {
		if (scheduler == null || completionQueue == null)
			throw new NullPointerException();
		this.scheduler = scheduler;
		this.completionQueue = completionQueue;
	}

	public Future<RES> submit(Callable<RES> task) {
		if (task == null)
			throw new NullPointerException();
		ScheduledQueueingFuture f = new ScheduledQueueingFuture(task);
		scheduler.execute(f);
		return f;
	}

	public Future<RES> submit(Runnable task, RES result) {
		if (task == null)
			throw new NullPointerException();
		ScheduledQueueingFuture f = new ScheduledQueueingFuture(task, result);
		scheduler.execute(f);
		return f;
	}

	/**
	 * Creates and executes a one-shot action that becomes enabled after the
	 * given delay.
	 * 
	 * @param command
	 *            the task to execute.
	 * @param delay
	 *            the time from now to delay execution.
	 * @param unit
	 *            the time unit of the delay parameter.
	 * @return a Future representing pending completion of the task, and whose
	 *         <tt>get()</tt> method will return <tt>null</tt> upon
	 *         completion.
	 * @throws RejectedExecutionException
	 *             if task cannot be scheduled for execution.
	 * @throws NullPointerException
	 *             if command is null
	 */
	public ScheduledFuture<?> schedule(Runnable command, long delay,
			TimeUnit unit) {
		if (command == null)
			throw new NullPointerException();
		ScheduledQueueingFuture f = new ScheduledQueueingFuture(command, null);
		return scheduler.schedule(f, delay, unit);
	}

	/**
	 * Creates and executes a ScheduledFuture that becomes enabled after the
	 * given delay.
	 * 
	 * @param callable
	 *            the function to execute.
	 * @param delay
	 *            the time from now to delay execution.
	 * @param unit
	 *            the time unit of the delay parameter.
	 * @return a ScheduledFuture that can be used to extract result or cancel.
	 * @throws RejectedExecutionException
	 *             if task cannot be scheduled for execution.
	 * @throws NullPointerException
	 *             if callable is null
	 */
	public ScheduledFuture<RES> schedule(Callable<RES> callable, long delay,
			TimeUnit unit) {
		if (callable == null)
			throw new NullPointerException();
		ScheduledQueueingFuture f = new ScheduledQueueingFuture(callable);
		ScheduledFuture<?> sf = scheduler.schedule(f, delay, unit);
		f.setDelay(sf.getDelay(TimeUnit.NANOSECONDS));

		return f;
	}

	public Future<RES> take() throws InterruptedException {
		return completionQueue.take();
	}

	public Future<RES> poll() {
		return completionQueue.poll();
	}

	public Future<RES> poll(long timeout, TimeUnit unit)
			throws InterruptedException {
		return completionQueue.poll(timeout, unit);
	}
}
